# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/scene.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12protos/scene.proto\x12\x04mira\"\x1d\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x05\x12\t\n\x01y\x18\x02 \x01(\x05\"\x9d\x01\n\nAnnotation\x12\x10\n\x08\x63\x61tegory\x18\x01 \x01(\x05\x12\n\n\x02x1\x18\x02 \x01(\x05\x12\n\n\x02x2\x18\x03 \x01(\x05\x12\n\n\x02y1\x18\x04 \x01(\x05\x12\n\n\x02y2\x18\x05 \x01(\x05\x12\x0f\n\x07is_rect\x18\x06 \x01(\x08\x12\x1b\n\x06points\x18\x07 \x03(\x0b\x32\x0b.mira.Point\x12\r\n\x05score\x18\x08 \x01(\x02\x12\x10\n\x08metadata\x18\t \x01(\t\"-\n\x17\x41nnotationConfiguration\x12\x12\n\ncategories\x18\x01 \x03(\t\"C\n\x04Mask\x12\x0f\n\x07visible\x18\x01 \x01(\x08\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1c\n\x07\x63ontour\x18\x03 \x03(\x0b\x32\x0b.mira.Point\"\xa4\x01\n\x05Scene\x12\x38\n\x11\x61nnotation_config\x18\x01 \x01(\x0b\x32\x1d.mira.AnnotationConfiguration\x12%\n\x0b\x61nnotations\x18\x02 \x03(\x0b\x32\x10.mira.Annotation\x12\x19\n\x05masks\x18\x03 \x03(\x0b\x32\n.mira.Mask\x12\r\n\x05image\x18\x04 \x01(\x0c\x12\x10\n\x08metadata\x18\x05 \x01(\tb\x06proto3')



_POINT = DESCRIPTOR.message_types_by_name['Point']
_ANNOTATION = DESCRIPTOR.message_types_by_name['Annotation']
_ANNOTATIONCONFIGURATION = DESCRIPTOR.message_types_by_name['AnnotationConfiguration']
_MASK = DESCRIPTOR.message_types_by_name['Mask']
_SCENE = DESCRIPTOR.message_types_by_name['Scene']
Point = _reflection.GeneratedProtocolMessageType('Point', (_message.Message,), {
  'DESCRIPTOR' : _POINT,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Point)
  })
_sym_db.RegisterMessage(Point)

Annotation = _reflection.GeneratedProtocolMessageType('Annotation', (_message.Message,), {
  'DESCRIPTOR' : _ANNOTATION,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Annotation)
  })
_sym_db.RegisterMessage(Annotation)

AnnotationConfiguration = _reflection.GeneratedProtocolMessageType('AnnotationConfiguration', (_message.Message,), {
  'DESCRIPTOR' : _ANNOTATIONCONFIGURATION,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.AnnotationConfiguration)
  })
_sym_db.RegisterMessage(AnnotationConfiguration)

Mask = _reflection.GeneratedProtocolMessageType('Mask', (_message.Message,), {
  'DESCRIPTOR' : _MASK,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Mask)
  })
_sym_db.RegisterMessage(Mask)

Scene = _reflection.GeneratedProtocolMessageType('Scene', (_message.Message,), {
  'DESCRIPTOR' : _SCENE,
  '__module__' : 'protos.scene_pb2'
  # @@protoc_insertion_point(class_scope:mira.Scene)
  })
_sym_db.RegisterMessage(Scene)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _POINT._serialized_start=28
  _POINT._serialized_end=57
  _ANNOTATION._serialized_start=60
  _ANNOTATION._serialized_end=217
  _ANNOTATIONCONFIGURATION._serialized_start=219
  _ANNOTATIONCONFIGURATION._serialized_end=264
  _MASK._serialized_start=266
  _MASK._serialized_end=333
  _SCENE._serialized_start=336
  _SCENE._serialized_end=500
# @@protoc_insertion_point(module_scope)
