from .preloaded import (  # noqa: F401
    load_voc2012,
    load_oxfordiiitpets,
    load_shapes,
    load_random_images,
    COCOAnnotationConfiguration,
    VOCAnnotationConfiguration,
    COCOAnnotationConfiguration90,
)
from .coco import load_coco  # noqa: F401
from .voc import load_voc  # noqa: F401
